"""Shared utilities and models for asynchronous and synchronous operations."""

import asyncio
import base64
import json
import sys
import threading
from abc import abstractmethod
from collections.abc import Coroutine
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from typing import Annotated, Any, Generic, TypeVar

# typing.Self is introduced in Python 3.11
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# typing.override is introduced in Python 3.12
if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from azure.core.credentials import AccessToken
from pydantic import (
    BaseModel,
    Field,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    model_validator,
)

R = TypeVar("R")
SP = TypeVar("SP", bound="SearchParams")
TOKEN_CREDENTIAL_SCOPE = "https://ossrdbms-aad.database.windows.net/.default"


class SSLMode(str, Enum):
    """SSL mode for Azure Database for PostgreSQL connections."""

    disable = "disable"
    allow = "allow"
    prefer = "prefer"
    require = "require"
    verify_ca = "verify-ca"
    verify_full = "verify-full"


class BasicAuth(BaseModel):
    """Basic username/password authentication for Azure Database for PostgreSQL connections.

    :param username: Username for the connection.
    :type username: str
    :param password: Password for the connection.
    :type password: str
    """

    username: str = "postgres"
    password: str = ""


class BaseConnectionInfo(BaseModel):
    """Base connection information for Azure Database for PostgreSQL connections.

    :param application_name: Name of the application connecting to the database.
    :type application_name: str
    :param host: Hostname of the Azure Database for PostgreSQL server.
    :type host: str | None
    :param dbname: Name of the database to connect to.
    :type dbname: str
    :param port: Port number for the connection.
    :type port: int
    :param sslmode: SSL mode for the connection.
    :type sslmode: SSLMode
    """

    application_name: str = "azure-postgresql"
    host: str | None = None
    dbname: str = "postgres"
    port: Annotated[NonNegativeInt, Field(le=65535)] = 5432
    sslmode: SSLMode = SSLMode.require


def run_coroutine_in_sync(coroutine: Coroutine[Any, Any, R]) -> R:
    def run_in_new_loop() -> R:
        new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(new_loop)
        try:
            return new_loop.run_until_complete(coroutine)
        finally:
            new_loop.close()

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        result = asyncio.run(coroutine)
    else:
        if threading.current_thread() is threading.main_thread():
            if not loop.is_running():
                result = loop.run_until_complete(coroutine)
            else:
                with ThreadPoolExecutor() as pool:
                    future = pool.submit(run_in_new_loop)
                    result = future.result()
        else:
            future = asyncio.run_coroutine_threadsafe(coroutine, loop)
            result = future.result()

    return result


def get_username_password(
    credentials: BasicAuth | AccessToken,
) -> tuple[str, str]:
    """Get username and password from credentials.

    :param credentials: BasicAuth for username/password or AccessToken for JWT token.
    :type credentials: BasicAuth | AccessToken
    :raises ValueError: User name not found in JWT token header
    :raises TypeError: Invalid credentials type
    :return: Tuple of username and password strings (plaintext).
    :rtype: tuple[str, str]
    """
    if isinstance(credentials, BasicAuth):
        return credentials.username, credentials.password
    elif isinstance(credentials, AccessToken):
        token = credentials.token
        _header, body_, _signature = token.split(".")
        body = json.loads(
            base64.b64decode(body_ + "=" * (4 - len(body_) % 4)).decode("utf-8")
        )
        username: str | None = body.get("upn", body.get("unique_name"))
        if username is None:
            raise ValueError("User name not found in JWT token header")
        return username, token
    else:
        raise TypeError(
            f"Invalid credentials type: {type(credentials)}. "
            "Expected BasicAuth or TokenCredential."
        )


class VectorOpClass(str, Enum):
    """Enumeration for operator classes used in vector indexes."""

    # Full-precision dense vector operator classes
    vector_cosine_ops = "vector_cosine_ops"
    vector_ip_ops = "vector_ip_ops"
    vector_l1_ops = "vector_l1_ops"
    vector_l2_ops = "vector_l2_ops"
    # Half-precision dense vector operator classes
    halfvec_cosine_ops = "halfvec_cosine_ops"
    halfvec_ip_ops = "halfvec_ip_ops"
    halfvec_l1_ops = "halfvec_l1_ops"
    halfvec_l2_ops = "halfvec_l2_ops"
    # Sparse vector operator classes
    sparsevec_cosine_ops = "sparsevec_cosine_ops"
    sparsevec_ip_ops = "sparsevec_ip_ops"
    sparsevec_l1_ops = "sparsevec_l1_ops"
    sparsevec_l2_ops = "sparsevec_l2_ops"
    # Bit vector operator classes
    bit_hamming_ops = "bit_hamming_ops"
    bit_jaccard_ops = "bit_jaccard_ops"

    def to_operator(self) -> str:
        """Return the distance operator as a string.

        :return: The distance operator string.
        :rtype: str
        :raises ValueError: If the vector operator class is unsupported.
        """
        match self:
            case (
                VectorOpClass.vector_cosine_ops
                | VectorOpClass.halfvec_cosine_ops
                | VectorOpClass.sparsevec_cosine_ops
            ):
                return "<=>"
            case (
                VectorOpClass.vector_ip_ops
                | VectorOpClass.halfvec_ip_ops
                | VectorOpClass.sparsevec_ip_ops
            ):
                return "<#>"
            case (
                VectorOpClass.vector_l1_ops
                | VectorOpClass.halfvec_l1_ops
                | VectorOpClass.sparsevec_l1_ops
            ):
                return "<+>"
            case (
                VectorOpClass.vector_l2_ops
                | VectorOpClass.halfvec_l2_ops
                | VectorOpClass.sparsevec_l2_ops
            ):
                return "<->"
            case VectorOpClass.bit_hamming_ops:
                return "<~>"
            case VectorOpClass.bit_jaccard_ops:
                return "<%>"
            case _:
                raise ValueError(f"Unsupported vector operator class: {self}")


class VectorType(str, Enum):
    """Enumeration for vector types used in vector similarity search."""

    bit = "bit"
    halfvec = "halfvec"
    sparsevec = "sparsevec"
    vector = "vector"


class Algorithm(BaseModel, Generic[SP]):
    op_class: VectorOpClass = VectorOpClass.vector_cosine_ops
    maintenance_work_mem: str | None = None
    max_parallel_maintenance_workers: Annotated[
        NonNegativeInt | None, Field(le=1_024)
    ] = None
    max_parallel_workers: Annotated[NonNegativeInt | None, Field(le=1_024)] = None

    @abstractmethod
    def default_search_params(self) -> SP:
        """Return the default search parameters for the algorithm.

        :return: An instance of the search parameters model.
        :rtype: SP
        """
        ...

    @abstractmethod
    def build_settings(self, exclude_none: bool = True) -> dict[str, Any]:
        """Return the specific index build settings for the algorithm.

        :param exclude_none: Whether to exclude keys with None values in the dictionary.
        :type exclude_none: bool
        :return: A dictionary containing the settings.
        :rtype: dict[str, Any]
        """
        ...

    def index_settings(self, exclude_none: bool = True) -> dict[str, Any]:
        """Return the general index settings for the algorithm.

        :param exclude_none: Whether to exclude keys with None values in the dictionary.
        :type exclude_none: bool
        :return: A dictionary containing the index settings.
        :rtype: dict[str, Any]
        """
        return {
            key: value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
            if key
            in [
                "maintenance_work_mem",
                "max_parallel_maintenance_workers",
                "max_parallel_workers",
            ]
        }


class SearchParams(BaseModel):
    @abstractmethod
    def search_settings(self, exclude_none: bool = True) -> dict[str, Any]:
        """Return the specific index search settings for the algorithm.

        :param exclude_none: Whether to exclude keys with None values in the dictionary.
        :type exclude_none: bool
        :return: A dictionary containing the search settings.
        :rtype: dict[str, Any]
        """
        ...


class DiskANNIterativeScanMode(str, Enum):
    """Enumeration for DiskANN iterative scan modes."""

    off = "off"
    relaxed = "relaxed_order"
    strict = "strict_order"


class DiskANNSearchParams(SearchParams):
    l_value_is: Annotated[PositiveInt | None, Field(ge=10, le=10_000)] = None
    iterative_search: DiskANNIterativeScanMode | None = None

    @override
    def search_settings(self, exclude_none=True):
        return {
            f"diskann.{key}": value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
        }


class DiskANN(Algorithm[DiskANNSearchParams]):
    max_neighbors: Annotated[PositiveInt | None, Field(ge=20, le=1_538)] = None
    l_value_ib: Annotated[PositiveInt | None, Field(ge=10, le=500)] = None
    product_quantized: bool | None = None
    pq_param_num_chunks: Annotated[NonNegativeInt | None, Field(le=8_000)] = None
    pq_param_training_samples: Annotated[NonNegativeInt | None, Field(le=1_000_000)] = (
        None
    )

    @model_validator(mode="after")
    def sanity_check(self) -> Self:
        if not self.product_quantized and self.pq_param_num_chunks is not None:
            raise ValueError(
                "Parameter 'product_quantized' must be True when 'pq_param_num_chunks' is set."
            )
        if not self.product_quantized and self.pq_param_training_samples is not None:
            raise ValueError(
                "Parameter 'product_quantized' must be True when 'pq_param_training_samples' is set."
            )
        return self

    @override
    def build_settings(self, exclude_none: bool = True) -> dict[str, Any]:
        return {
            key: value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
            if key
            in [
                "max_neighbors",
                "l_value_ib",
                "product_quantized",
                "pq_param_num_chunks",
                "pq_param_training_samples",
            ]
        }

    @override
    def default_search_params(self) -> DiskANNSearchParams:
        return DiskANNSearchParams()


class HNSWIterativeScanMode(str, Enum):
    """Enumeration for HNSW iterative scan modes."""

    off = "off"
    relaxed = "relaxed_order"
    strict = "strict_order"


class HNSWSearchParams(SearchParams):
    ef_search: Annotated[PositiveInt | None, Field(le=1_000)] = None
    iterative_scan: HNSWIterativeScanMode | None = None
    max_scan_tuples: PositiveInt | None = None
    scan_mem_multiplier: Annotated[PositiveFloat | None, Field(le=1_000)] = None

    @override
    def search_settings(self, exclude_none=True):
        return {
            f"hnsw.{key}": value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
        }


class HNSW(Algorithm[HNSWSearchParams]):
    m: Annotated[PositiveInt | None, Field(ge=2, le=100)] = None
    ef_construction: Annotated[PositiveInt | None, Field(ge=4, le=1_000)] = None

    @model_validator(mode="after")
    def sanity_check(self) -> Self:
        if (
            self.m is not None
            and self.ef_construction is not None
            and self.ef_construction < 2 * self.m
        ):
            raise ValueError(
                "Parameter 'ef_construction' must be at least twice the value of 'm'."
            )
        return self

    @override
    def build_settings(self, exclude_none=True):
        return {
            key: value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
            if key in ["m", "ef_construction"]
        }

    @override
    def default_search_params(self) -> HNSWSearchParams:
        return HNSWSearchParams()


class IVFFlatIterativeScanMode(str, Enum):
    """Enumeration for IVFFlat iterative scan modes."""

    off = "off"
    relaxed = "relaxed_order"


class IVFFlatSearchParams(SearchParams):
    probes: Annotated[PositiveInt | None, Field(le=32_768)] = None
    iterative_scan: IVFFlatIterativeScanMode | None = None
    max_probes: Annotated[PositiveInt | None, Field(le=32_768)] = None

    @override
    def search_settings(self, exclude_none=True):
        return {
            f"ivfflat.{key}": value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
        }


class IVFFlat(Algorithm[IVFFlatSearchParams]):
    lists: Annotated[PositiveInt | None, Field(le=32_768)] = None

    @override
    def build_settings(self, exclude_none: bool = True) -> dict[str, Any]:
        return {
            key: value
            for key, value in self.model_dump(
                mode="json", exclude_none=exclude_none
            ).items()
            if key in ["lists"]
        }

    @override
    def default_search_params(self) -> IVFFlatSearchParams:
        return IVFFlatSearchParams()


class Extension(BaseModel):
    """Model representing a PostgreSQL extension.

    :param ext_name: Name of the extension to be created, checked or dropped.
    :type ext_name: str
    :param ext_version: Optional version of the extension to be created or checked.
    :type ext_version: str | None
    :param schema_name: Optional schema name where the extension should be created
                        or checked.
    :type schema_name: str | None
    :param cascade: Whether to automatically install the extension dependencies or
                    drop the objects that depend on the extension.
    :type cascade: bool
    """

    ext_name: str
    ext_version: str | None = None
    schema_name: str | None = None
    cascade: bool = False



"""Asynchronous connection handling for Azure Database for PostgreSQL."""

import logging
import time
from collections.abc import Awaitable, Callable

from azure.core.credentials_async import AsyncTokenCredential
from azure.identity.aio import DefaultAzureCredential
from psycopg import AsyncConnection, sql
from psycopg.rows import dict_row
from psycopg_pool import AsyncConnectionPool
from pydantic import ConfigDict

from .._shared import (
    TOKEN_CREDENTIAL_SCOPE,
    BaseConnectionInfo,
    BasicAuth,
    Extension,
    get_username_password,
    run_coroutine_in_sync,
)

_logger = logging.getLogger(__name__)


async def async_check_connection(
    conn: AsyncConnection, /, required_extensions: list[Extension] = []
):
    """Check if the connection to Azure Database for PostgreSQL is valid and required extensions are installed.

    :param conn: Async connection to the Azure Database for PostgreSQL.
    :type conn: AsyncConnection
    :param required_extensions: List of required extensions to check if they are installed.
    :type required_extensions: list[Extension]
    :raises RuntimeError: If the connection check fails or required extensions are not installed.
    """
    async with conn.cursor(row_factory=dict_row) as cursor:
        _logger.debug("checking connection")
        t_start = time.perf_counter()
        await cursor.execute("select 1")
        result = await cursor.fetchone()
        t_elapsed = time.perf_counter() - t_start
        assert result is not None, "Connection check failed: no result returned."
        _logger.debug(
            "connection check successful. elapsed time: %.3f ms", t_elapsed * 1000
        )

        for ext in required_extensions:
            ext_name = ext.ext_name
            ext_version = ext.ext_version
            schema_name = ext.schema_name
            await cursor.execute(
                sql.SQL(
                    """
                    select  extname as ext_name, extversion as ext_version,
                            n.nspname as schema_name
                      from  pg_extension e
                            left join pg_namespace n on e.extnamespace = n.oid
                     where  extname = %(ext_name)s
                    """
                ),
                {"ext_name": ext_name},
            )
            resultset = await cursor.fetchone()
            if resultset is None:
                raise RuntimeError(f"Required extension '{ext_name}' is not installed.")
            if ext_version is not None and resultset["ext_version"] != ext_version:
                raise RuntimeError(
                    f"Required extension '{ext_name}' version mismatch: "
                    f"expected {ext_version}, got {resultset['ext_version']}."
                )
            if schema_name is not None and resultset["schema_name"] != schema_name:
                raise RuntimeError(
                    f"Required extension '{ext_name}' is not installed in the expected schema: "
                    f"expected {schema_name}, got {resultset['schema_name']}."
                )
            _logger.debug(
                "required extension '%s' is installed (version: %s, schema: %s)",
                resultset["ext_name"],
                resultset["ext_version"],
                resultset["schema_name"],
            )


async def async_create_extensions(
    conn: AsyncConnection, /, required_extensions: list[Extension] = []
):
    """Create required extensions in the Azure Database for PostgreSQL connection.

    :param conn: Async connection to the Azure Database for PostgreSQL.
    :type conn: AsyncConnection
    :param required_extensions: List of required extensions to create.
    :type required_extensions: list[Extension]
    :raises Exception: If the connection is not valid or if an error occurs during extension creation.
    """
    async with conn.cursor() as cursor:
        for ext in required_extensions:
            ext_name = ext.ext_name
            ext_version = ext.ext_version
            schema_name = ext.schema_name
            cascade = ext.cascade
            _logger.debug(
                "creating extension (if not exists): %s (version: %s, schema: %s, cascade: %s)",
                ext_name,
                ext_version,
                schema_name,
                cascade,
            )
            await cursor.execute(
                sql.SQL(
                    """
                    create extension  if not exists {ext_name}
                                with  {schema_expr}
                                      {version_expr}
                                      {cascade_expr}
                    """
                ).format(
                    ext_name=sql.Identifier(ext_name),
                    schema_expr=sql.SQL("schema {schema_name}").format(
                        schema_name=sql.Identifier(schema_name)
                    )
                    if schema_name is not None
                    else sql.SQL(""),
                    version_expr=sql.SQL("version {version}").format(
                        version=sql.Literal(ext_version)
                    )
                    if ext_version is not None
                    else sql.SQL(""),
                    cascade_expr=sql.SQL("cascade") if cascade else sql.SQL(""),
                )
            )


class AsyncConnectionInfo(BaseConnectionInfo):
    """Base connection information for Azure Database for PostgreSQL connections.

    :param host: Hostname of the Azure Database for PostgreSQL server.
    :type host: str | None
    :param dbname: Name of the database to connect to.
    :type dbname: str
    :param port: Port number for the connection.
    :type port: int
    :param credentials: Credentials for authentication.
    :type credentials: BasicAuth | AsyncTokenCredential
    :param sslmode: SSL mode for the connection.
    :type sslmode: SSLMode
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,  # True to allow AsyncTokenCredential
    )

    credentials: BasicAuth | AsyncTokenCredential = DefaultAzureCredential()


class AsyncAzurePGConnectionPool(AsyncConnectionPool):
    """Async connection pool for Azure Database for PostgreSQL connections."""

    def __init__(
        self,
        conninfo: str = "",
        *,
        azure_conn_info: AsyncConnectionInfo = AsyncConnectionInfo(),
        **kwargs,
    ):
        if isinstance(azure_conn_info.credentials, AsyncTokenCredential):
            credential_provider = azure_conn_info.credentials
            coroutine = credential_provider.get_token(TOKEN_CREDENTIAL_SCOPE)

            _logger.debug(
                "getting token from TokenCredential for the scope: %s",
                TOKEN_CREDENTIAL_SCOPE,
            )

            token = run_coroutine_in_sync(coroutine)

            _logger.info("getting username and password from token")
            username, password = get_username_password(token)

            _logger.debug("wrapping reconnect_failed function")
            reconnect_failed: (
                Callable[[AsyncConnectionPool], Awaitable[None]] | None
            ) = kwargs.get("reconnect_failed")

            async def reconnect_failed_wrapper(pool: AsyncConnectionPool) -> None:
                if reconnect_failed:
                    await reconnect_failed(pool)

                _logger.debug(
                    "getting token from TokenCredential for the scope: %s",
                    TOKEN_CREDENTIAL_SCOPE,
                )
                token = await credential_provider.get_token(TOKEN_CREDENTIAL_SCOPE)

                _logger.info("getting username and password from token")
                username, password = get_username_password(token)

                pool.kwargs.update(
                    user=username,
                    password=password,
                )

            kwargs["reconnect_failed"] = reconnect_failed_wrapper
        else:
            username, password = get_username_password(azure_conn_info.credentials)

        azure_conn_info_kwargs = azure_conn_info.model_dump(
            mode="json", exclude_none=True, exclude=set(["credentials"])
        )
        _logger.debug(
            "updating AsyncConnectionPool kwargs with those from: %s",
            azure_conn_info_kwargs,
        )
        kwargs_ = kwargs.get("kwargs", {})
        kwargs_.update(user=username, password=password, **azure_conn_info_kwargs)
        kwargs["kwargs"] = kwargs_

        super().__init__(conninfo, **kwargs)



"""Base VectorStore integration for Azure Database for PostgreSQL."""

import logging
import re
import sys
from collections.abc import Sequence
from typing import Any

import numpy as np
from pgvector.psycopg import register_vector  # type: ignore[import-untyped]
from psycopg import sql
from psycopg.rows import dict_row
from pydantic import BaseModel, ConfigDict, PositiveInt, model_validator

from ._connection import AzurePGConnectionPool
from ._shared import (
    HNSW,
    Algorithm,
    DiskANN,
    IVFFlat,
    VectorOpClass,
    VectorType,
)

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

_logger = logging.getLogger(__name__)


class BaseAzurePGVectorStore(BaseModel):
    """Base Pydantic model for an Azure PostgreSQL-backed vector store.

    This class encapsulates configuration (connection pool, table/column
    names, embedding type/dimension, index configuration and metadata
    column) and performs runtime verification that the target table
    exists with expected columns and index configuration. If the table
    does not exist, ``verify_and_init_store`` will create it.
    """

    connection_pool: AzurePGConnectionPool
    schema_name: str = "public"
    table_name: str = "vector_store"
    id_column: str = "id"
    content_column: str = "content"
    embedding_column: str = "embedding"
    embedding_type: VectorType | None = None
    embedding_dimension: PositiveInt | None = None
    embedding_index: Algorithm | None = None
    metadata_column: str | None = "metadata"

    model_config = ConfigDict(
        arbitrary_types_allowed=True,  # Allow arbitrary types like Embeddings and AzurePGConnectionPool
    )

    @model_validator(mode="after")
    def verify_and_init_store(self) -> Self:
        """Validate the store configuration and initialize DB schema and index.

        This validator runs after Pydantic model initialization. It queries
        the database to detect an existing table and its columns/indexes,
        performs type and dimension checks for the embedding column, and
        sets inferred properties (like embedding_type and embedding_dimension)
        when they are not explicitly provided. If the table does not exist,
        it will create the table with sensible defaults.

        Returns:
            Self: The same model instance, possibly updated with inferred values.
        """
        # verify that metadata_column is not empty if provided
        if self.metadata_column is not None and len(self.metadata_column) == 0:
            raise ValueError("'metadata_column' cannot be empty if provided.")

        _logger.debug(
            "checking if table '%s.%s' exists with the required columns",
            self.schema_name,
            self.table_name,
        )

        with (
            self.connection_pool.connection() as conn,
            conn.cursor(row_factory=dict_row) as cursor,
        ):
            cursor.execute(
                sql.SQL(
                    """
                      select  a.attname as column_name,
                              format_type(a.atttypid, a.atttypmod) as column_type
                        from  pg_attribute a
                              join pg_class c on a.attrelid = c.oid
                              join pg_namespace n on c.relnamespace = n.oid
                       where  a.attnum > 0
                              and not a.attisdropped
                              and n.nspname = %(schema_name)s
                              and c.relname = %(table_name)s
                    order by  a.attnum asc
                    """
                ),
                {"schema_name": self.schema_name, "table_name": self.table_name},
            )
            resultset = cursor.fetchall()
            existing_columns: dict[str, str] = {
                row["column_name"]: row["column_type"] for row in resultset
            }

        # if table exists, verify that required columns exist and have correct types
        if len(existing_columns) > 0:
            _logger.debug(
                "table '%s.%s' exists with the following column mapping: %s",
                self.schema_name,
                self.table_name,
                existing_columns,
            )

            id_column_type = existing_columns.get(self.id_column)
            if id_column_type != "uuid":
                raise ValueError(
                    f"Table '{self.schema_name}.{self.table_name}' must have a column '{self.id_column}' of type 'uuid'."
                )

            content_column_type = existing_columns.get(self.content_column)
            if content_column_type is None or (
                content_column_type != "text"
                and not content_column_type.startswith("varchar")
            ):
                raise ValueError(
                    f"Table '{self.schema_name}.{self.table_name}' must have a column '{self.content_column}' of type 'text' or 'varchar'."
                )

            embedding_column_type = existing_columns.get(self.embedding_column)
            pattern = re.compile(r"(?P<type>\w+)(?:\((?P<dim>\d+)\))?")
            m = pattern.match(embedding_column_type if embedding_column_type else "")
            parsed_type: str | None = m.group("type") if m else None
            parsed_dim: PositiveInt | None = (
                PositiveInt(m.group("dim")) if m and m.group("dim") else None
            )

            vector_types = [t.value for t in VectorType.__members__.values()]
            if parsed_type not in vector_types:
                raise ValueError(
                    f"Column '{self.embedding_column}' in table '{self.schema_name}.{self.table_name}' must be one of the following types: {vector_types}."
                )
            elif (
                self.embedding_type is not None
                and parsed_type != self.embedding_type.value
            ):
                raise ValueError(
                    f"Column '{self.embedding_column}' in table '{self.schema_name}.{self.table_name}' has type '{parsed_type}', but the specified embedding_type is '{self.embedding_type.value}'. They must match."
                )
            elif self.embedding_type is None:
                _logger.info(
                    "embedding_type is not specified, but the column '%s' in table '%s.%s' has type '%s'. Overriding embedding_type accordingly.",
                    self.embedding_column,
                    self.schema_name,
                    self.table_name,
                    parsed_type,
                )
                self.embedding_type = VectorType(parsed_type)

            if parsed_dim is not None and self.embedding_dimension is None:
                _logger.info(
                    "embedding_dimension is not specified, but the column '%s' in table '%s.%s' has a dimension of %d. Overriding embedding_dimension accordingly.",
                    self.embedding_column,
                    self.schema_name,
                    self.table_name,
                    parsed_dim,
                )
                self.embedding_dimension = parsed_dim
            elif (
                parsed_dim is not None
                and self.embedding_dimension is not None
                and parsed_dim != self.embedding_dimension
            ):
                raise ValueError(
                    f"Column '{self.embedding_column}' in table '{self.schema_name}.{self.table_name}' has a dimension of {parsed_dim}, but the specified embedding_dimension is {self.embedding_dimension}. They must match."
                )

            if self.metadata_column is not None:
                existing_type = existing_columns.get(self.metadata_column)
                if existing_type is None:
                    raise ValueError(
                        f"Column '{self.metadata_column}' does not exist in table '{self.schema_name}.{self.table_name}'."
                    )

            with (
                self.connection_pool.connection() as conn,
                conn.cursor(row_factory=dict_row) as cursor,
            ):
                _logger.debug(
                    "checking if table '%s.%s' has a vector index on column '%s'",
                    self.schema_name,
                    self.table_name,
                    self.embedding_column,
                )
                cursor.execute(
                    sql.SQL(
                        """
                        with cte as (
                          select  n.nspname as schema_name,
                                  ct.relname as table_name,
                                  ci.relname as index_name,
                                  a.amname as index_type,
                                  pg_get_indexdef(
                                    ci.oid, -- index OID
                                    generate_series(1, array_length(ii.indkey, 1)), -- column no
                                    true -- pretty print
                                  ) as index_column,
                                  o.opcname as index_opclass,
                                  ci.reloptions as index_opts
                            from  pg_class ci
                                  join pg_index ii on ii.indexrelid = ci.oid
                                  join pg_am a on a.oid = ci.relam
                                  join pg_class ct on ct.oid = ii.indrelid
                                  join pg_namespace n on n.oid = ci.relnamespace
                                  join pg_opclass o on o.oid = any(ii.indclass)
                           where  ci.relkind = 'i'
                                  and ct.relkind = 'r'
                                  and ii.indisvalid
                                  and ii.indisready
                        ) select  schema_name, table_name, index_name, index_type,
                                  index_column, index_opclass, index_opts
                            from  cte
                           where  schema_name = %(schema_name)s
                                  and table_name = %(table_name)s
                                  and index_column like %(embedding_column)s
                                  and (
                                      index_opclass like '%%vector%%'
                                      or index_opclass like '%%halfvec%%'
                                      or index_opclass like '%%sparsevec%%'
                                      or index_opclass like '%%bit%%'
                                  )
                        order by  schema_name, table_name, index_name
                        """
                    ),
                    {
                        "schema_name": self.schema_name,
                        "table_name": self.table_name,
                        "embedding_column": f"%{self.embedding_column}%",
                    },
                )
                resultset = cursor.fetchall()

            if len(resultset) > 0:
                _logger.debug(
                    "table '%s.%s' has %d vector index(es): %s",
                    self.schema_name,
                    self.table_name,
                    len(resultset),
                    resultset,
                )

                if self.embedding_index is None:
                    _logger.info(
                        "embedding_index is not specified, using the first found index: %s",
                        resultset[0],
                    )

                    index_type = resultset[0]["index_type"]
                    index_opclass = VectorOpClass(resultset[0]["index_opclass"])
                    index_opts = {
                        opts.split("=")[0]: opts.split("=")[1]
                        for opts in resultset[0]["index_opts"]
                    }

                    index = (
                        DiskANN(op_class=index_opclass, **index_opts)
                        if index_type == "diskann"
                        else (
                            HNSW(op_class=index_opclass, **index_opts)
                            if index_type == "hnsw"
                            else IVFFlat(op_class=index_opclass, **index_opts)
                        )
                    )

                    self.embedding_index = index
                else:
                    _logger.info(
                        "embedding_index is specified as '%s'; will try to find a matching index.",
                        self.embedding_index,
                    )
                    print(resultset)
                    index_opclass = self.embedding_index.op_class.value  # type: ignore[assignment]
                    if isinstance(self.embedding_index, DiskANN):
                        index_type = "diskann"
                    elif isinstance(self.embedding_index, HNSW):
                        index_type = "hnsw"
                    else:
                        index_type = "ivfflat"

                    for row in resultset:
                        if (
                            row["index_type"] == index_type
                            and row["index_opclass"] == index_opclass
                        ):
                            _logger.info(
                                "found a matching index: %s. overriding embedding_index.",
                                row,
                            )
                            index_opts = {
                                opts.split("=")[0]: opts.split("=")[1]
                                for opts in row["index_opts"]
                            }
                            index = (
                                DiskANN(op_class=index_opclass, **index_opts)
                                if index_type == "diskann"
                                else (
                                    HNSW(op_class=index_opclass, **index_opts)
                                    if index_type == "hnsw"
                                    else IVFFlat(op_class=index_opclass, **index_opts)
                                )
                            )
                            self.embedding_index = index
                            break
            elif self.embedding_index is None:
                _logger.info(
                    "embedding_index is not specified, and no vector index found in table '%s.%s'. defaulting to 'DiskANN' with 'vector_cosine_ops' opclass.",
                    self.schema_name,
                    self.table_name,
                )
                self.embedding_index = DiskANN(op_class=VectorOpClass.vector_cosine_ops)

        # if table does not exist, create it
        else:
            _logger.debug(
                "table '%s.%s' does not exist, creating it with the required columns",
                self.schema_name,
                self.table_name,
            )

            if self.embedding_type is None:
                _logger.warning(
                    "Embedding type is not specified, defaulting to 'vector'."
                )
                self.embedding_type = VectorType.vector

            if self.embedding_dimension is None:
                _logger.warning(
                    "Embedding dimension is not specified, defaulting to 1536."
                )
                self.embedding_dimension = PositiveInt(1_536)

            if self.embedding_index is None:
                _logger.warning(
                    "Embedding index is not specified, defaulting to 'DiskANN' with 'vector_cosine_ops' opclass."
                )
                self.embedding_index = DiskANN(op_class=VectorOpClass.vector_cosine_ops)

            with self.connection_pool.connection() as conn, conn.cursor() as cursor:
                cursor.execute(
                    sql.SQL(
                        """
                        create table {table_name} (
                            {id_column} uuid primary key,
                            {content_column} text,
                            {embedding_column} {embedding_type}({embedding_dimension}),
                            {metadata_column} jsonb
                        )
                        """
                    ).format(
                        table_name=sql.Identifier(self.schema_name, self.table_name),
                        id_column=sql.Identifier(self.id_column),
                        content_column=sql.Identifier(self.content_column),
                        embedding_column=sql.Identifier(self.embedding_column),
                        embedding_type=sql.Identifier(self.embedding_type.value),
                        embedding_dimension=sql.Literal(self.embedding_dimension),
                        metadata_column=sql.Identifier(self.metadata_column),
                    )
                )

        return self

    def _delete_rows_from_table(
        self, ids: list[str] | None = None, **kwargs: Any
    ) -> bool | None:
        """Delete rows from the table by their IDs or truncate the table.

        Args:
            ids (list[str] | None): List of IDs to delete. If None, truncates the table.
            **kwargs: Additional options, such as 'restart' and 'cascade' for truncation.

        Returns:
            bool | None: True if successful, False if an exception occurred, None otherwise.
        """
        with self.connection_pool.connection() as conn:
            conn.autocommit = False
            try:
                with conn.transaction() as _tx, conn.cursor() as cursor:
                    if ids is None:
                        restart = bool(kwargs.pop("restart", None))
                        cascade = bool(kwargs.pop("cascade", None))
                        cursor.execute(
                            sql.SQL(
                                """
                                truncate table {table_name} {restart} {cascade}
                                """
                            ).format(
                                table_name=sql.Identifier(
                                    self.schema_name, self.table_name
                                ),
                                restart=sql.SQL(
                                    "restart identity"
                                    if restart
                                    else "continue identity"
                                ),
                                cascade=sql.SQL("cascade" if cascade else "restrict"),
                            )
                        )
                    else:
                        cursor.execute(
                            sql.SQL(
                                """
                                delete from {table_name}
                                      where {id_column} = any(%(id)s)
                                """
                            ).format(
                                table_name=sql.Identifier(
                                    self.schema_name, self.table_name
                                ),
                                id_column=sql.Identifier(self.id_column),
                            ),
                            {"id": ids},
                        )
            except Exception:
                return False
            else:
                return True

    def _similarity_search_by_vector_with_distance(
        self, embedding: list[float], k: int = 4, **kwargs: Any
    ) -> list[tuple[dict, float, np.ndarray | None]]:
        """Perform a similarity search using a vector embedding and return results with distances.

        Args:
            embedding (list[float]): The query embedding vector.
            k (int): Number of top results to return.
            **kwargs: Additional options such as 'return_embeddings', 'top_m', and 'filter_expression'.

        Returns:
            list[tuple[dict, float, np.ndarray | None]]: List of tuples containing document dict, distance, and optionally the embedding.
        """
        assert self.embedding_index is not None, (
            "embedding_index should have already been set"
        )
        return_embeddings = bool(kwargs.pop("return_embeddings", None))
        top_m = int(kwargs.pop("top_m", 5 * k))
        filter_expression: sql.SQL = kwargs.pop("filter_expression", sql.SQL("true"))
        with self.connection_pool.connection() as conn:
            register_vector(conn)
            with conn.cursor(row_factory=dict_row) as cursor:
                metadata_column: list[str]
                if isinstance(self.metadata_column, list):
                    metadata_column = [
                        col if isinstance(col, str) else col[0]
                        for col in self.metadata_column
                    ]
                elif isinstance(self.metadata_column, str):
                    metadata_column = [self.metadata_column]
                else:
                    metadata_column = []

                # do reranking for the following cases:
                #   - binary or scalar quantizations (for HNSW and IVFFlat), or
                #   - product quantization (for DiskANN)
                if (
                    self.embedding_index.op_class == VectorOpClass.bit_hamming_ops
                    or self.embedding_index.op_class == VectorOpClass.bit_jaccard_ops
                    or self.embedding_index.op_class == VectorOpClass.halfvec_cosine_ops
                    or self.embedding_index.op_class == VectorOpClass.halfvec_ip_ops
                    or self.embedding_index.op_class == VectorOpClass.halfvec_l1_ops
                    or self.embedding_index.op_class == VectorOpClass.halfvec_l2_ops
                    or (
                        isinstance(self.embedding_index, DiskANN)
                        and self.embedding_index.product_quantized
                    )
                ):
                    sql_query = sql.SQL(
                        """
                          select  {outer_columns},
                                  {embedding_column} {op} %(query)s as distance,
                                  {maybe_embedding_column}
                            from  (
                                     select {inner_columns}
                                       from {table_name}
                                      where {filter_expression}
                                   order by {expression} asc
                                      limit %(top_m)s
                                  ) i
                        order by  {embedding_column} {op} %(query)s asc
                           limit  %(top_k)s
                        """
                    ).format(
                        outer_columns=sql.SQL(", ").join(
                            map(
                                sql.Identifier,
                                [
                                    self.id_column,
                                    self.content_column,
                                    *metadata_column,
                                ],
                            )
                        ),
                        embedding_column=sql.Identifier(self.embedding_column),
                        op=(
                            sql.SQL(
                                VectorOpClass.vector_cosine_ops.to_operator()
                            )  # TODO(arda): Think of getting this from outside
                            if (
                                self.embedding_index.op_class
                                in (
                                    VectorOpClass.bit_hamming_ops,
                                    VectorOpClass.bit_jaccard_ops,
                                )
                            )
                            else sql.SQL(self.embedding_index.op_class.to_operator())
                        ),
                        maybe_embedding_column=(
                            sql.Identifier(self.embedding_column)
                            if return_embeddings
                            else sql.SQL(" as ").join(
                                (sql.NULL, sql.Identifier(self.embedding_column))
                            )
                        ),
                        inner_columns=sql.SQL(", ").join(
                            map(
                                sql.Identifier,
                                [
                                    self.id_column,
                                    self.content_column,
                                    self.embedding_column,
                                    *metadata_column,
                                ],
                            )
                        ),
                        table_name=sql.Identifier(self.schema_name, self.table_name),
                        filter_expression=filter_expression,
                        expression=(
                            sql.SQL(
                                "binary_quantize({embedding_column})::bit({embedding_dim}) {op} binary_quantize({query})"
                            ).format(
                                embedding_column=sql.Identifier(self.embedding_column),
                                embedding_dim=sql.Literal(self.embedding_dimension),
                                op=sql.SQL(self.embedding_index.op_class.to_operator()),
                                query=sql.Placeholder("query"),
                            )
                            if self.embedding_index.op_class
                            in (
                                VectorOpClass.bit_hamming_ops,
                                VectorOpClass.bit_jaccard_ops,
                            )
                            else (
                                sql.SQL(
                                    "{embedding_column}::halfvec({embedding_dim}) {op} {query}::halfvec({embedding_dim})"
                                ).format(
                                    embedding_column=sql.Identifier(
                                        self.embedding_column
                                    ),
                                    embedding_dim=sql.Literal(self.embedding_dimension),
                                    op=sql.SQL(
                                        self.embedding_index.op_class.to_operator()
                                    ),
                                    query=sql.Placeholder("query"),
                                )
                                if self.embedding_index.op_class
                                in (
                                    VectorOpClass.halfvec_cosine_ops,
                                    VectorOpClass.halfvec_ip_ops,
                                    VectorOpClass.halfvec_l1_ops,
                                    VectorOpClass.halfvec_l2_ops,
                                )
                                else sql.SQL("{embedding_column} {op} {query}").format(
                                    embedding_column=sql.Identifier(
                                        self.embedding_column
                                    ),
                                    op=sql.SQL(
                                        self.embedding_index.op_class.to_operator()
                                    ),
                                    query=sql.Placeholder("query"),
                                )
                            )
                        ),
                    )
                # otherwise (i.e., no quantization), do not do reranking
                else:
                    sql_query = sql.SQL(
                        """
                          select  {outer_columns},
                                  {embedding_column} {op} %(query)s as distance,
                                  {maybe_embedding_column}
                            from  {table_name}
                           where  {filter_expression}
                        order by  {embedding_column} {op} %(query)s asc
                           limit  %(top_k)s
                        """
                    ).format(
                        outer_columns=sql.SQL(", ").join(
                            map(
                                sql.Identifier,
                                [
                                    self.id_column,
                                    self.content_column,
                                    *metadata_column,
                                ],
                            )
                        ),
                        embedding_column=sql.Identifier(self.embedding_column),
                        op=sql.SQL(self.embedding_index.op_class.to_operator()),
                        maybe_embedding_column=(
                            sql.Identifier(self.embedding_column)
                            if return_embeddings
                            else sql.SQL(" as ").join(
                                (sql.NULL, sql.Identifier(self.embedding_column))
                            )
                        ),
                        table_name=sql.Identifier(self.schema_name, self.table_name),
                        filter_expression=filter_expression,
                    )
                cursor.execute(
                    sql_query,
                    {
                        "query": np.array(embedding, dtype=np.float32),
                        "top_m": top_m,
                        "top_k": k,
                    },
                )
                resultset = cursor.fetchall()
        return [
            (
                {
                    "id": result[self.id_column],
                    "content": result[self.content_column],
                    "metadata": (
                        result[metadata_column[0]]
                        if isinstance(self.metadata_column, str)
                        else {col: result[col] for col in metadata_column}
                    ),
                },
                result["distance"],
                result.get(self.embedding_column),  # type: ignore[return-value]
            )
            for result in resultset
        ]

    def _get_by_ids(self, ids: Sequence[str], /) -> list[dict[str, Any]]:
        """Retrieve documents from the table by their IDs.

        Args:
            ids (Sequence[str]): List of IDs to retrieve.

        Returns:
            list[dict[str, Any]]: List of document dictionaries with id, content, embedding, and metadata.
        """
        with (
            self.connection_pool.connection() as conn,
            conn.cursor(row_factory=dict_row) as cursor,
        ):
            metadata_column: list[str]
            if isinstance(self.metadata_column, list):
                metadata_column = [
                    col if isinstance(col, str) else col[0]
                    for col in self.metadata_column
                ]
            elif isinstance(self.metadata_column, str):
                metadata_column = [self.metadata_column]
            else:
                metadata_column = []

            if ids is not None:
                where_clause = sql.SQL(" where {id_column} = any(%(id)s)").format(
                    id_column=sql.Identifier(self.id_column)
                )
            else:
                where_clause = sql.SQL("")

            get_sql = sql.SQL(
                """
                select {columns}
                    from {table_name}
                {where_clause}
                """
            ).format(
                columns=sql.SQL(", ").join(
                    map(
                        sql.Identifier,
                        [
                            self.id_column,
                            self.content_column,
                            self.embedding_column,
                            *metadata_column,
                        ],
                    )
                ),
                table_name=sql.Identifier(self.schema_name, self.table_name),
                where_clause=where_clause,
            )

            if ids is not None:
                cursor.execute(get_sql, {"id": ids})
            else:
                cursor.execute(get_sql)
            resultset = cursor.fetchall()
            documents = [
                {
                    "id": result[self.id_column],
                    "content": result[self.content_column],
                    "embedding": result[self.embedding_column],
                    "metadata": (
                        result[metadata_column[0]]
                        if isinstance(self.metadata_column, str)
                        else {col: result[col] for col in metadata_column}
                    ),
                }
                for result in resultset
            ]
            return documents



"""VectorStore integration for Azure Database for PostgreSQL using LlamaIndex."""

import sys
from typing import Any, Optional, Union

import numpy as np
from pgvector.psycopg import register_vector  # type: ignore[import-untyped]
from psycopg import sql
from psycopg.rows import dict_row
from psycopg.types.json import Jsonb

from llama_index.core.schema import BaseNode, MetadataMode
from llama_index.core.vector_stores.types import (
    BasePydanticVectorStore,
    FilterOperator,
    MetadataFilter,
    MetadataFilters,
    VectorStoreQuery,
    VectorStoreQueryResult,
)
from llama_index.core.vector_stores.utils import (
    metadata_dict_to_node,
    node_to_metadata_dict,
)

from .common import (
    Algorithm,
    AzurePGConnectionPool,
    BaseAzurePGVectorStore,
)

if sys.version_info < (3, 12):
    from typing_extensions import override
else:
    from typing import override


def metadata_filters_to_sql(filters: Optional[MetadataFilters]) -> sql.SQL:
    """Convert LlamaIndex MetadataFilters to a SQL WHERE clause.

    Args:
        filters: Optional MetadataFilters object.

    Returns:
        sql.SQL: SQL WHERE clause representing the filters.
    """
    if not filters or not filters.filters:
        return sql.SQL("TRUE")

    def _filter_to_sql(filter_item: Union[MetadataFilter, MetadataFilters]) -> sql.SQL:
        """Recursively convert MetadataFilter or MetadataFilters to SQL."""
        if isinstance(filter_item, MetadataFilters):
            # Handle nested MetadataFilters
            if not filter_item.filters:
                return sql.SQL("TRUE")

            filter_sqls = [_filter_to_sql(f) for f in filter_item.filters]
            if filter_item.condition.lower() == "and":
                return sql.SQL("({})").format(sql.SQL(" AND ").join(filter_sqls))
            elif filter_item.condition.lower() == "or":
                return sql.SQL("({})").format(sql.SQL(" OR ").join(filter_sqls))
            else:  # NOT
                if len(filter_sqls) == 1:
                    return sql.SQL("NOT ({})").format(filter_sqls[0])
                else:
                    # For multiple filters with NOT, apply NOT to the AND of all filters
                    return sql.SQL("NOT ({})").format(
                        sql.SQL(" AND ").join(filter_sqls)
                    )

        elif isinstance(filter_item, MetadataFilter):
            # Handle individual MetadataFilter
            key = filter_item.key
            value = filter_item.value
            operator = filter_item.operator

            # Use JSONB path for metadata column
            column_ref = sql.SQL("metadata ->> {}").format(sql.Literal(key))

            if operator == FilterOperator.EQ:
                return sql.SQL("{} = {}").format(column_ref, sql.Literal(str(value)))
            elif operator == FilterOperator.NE:
                return sql.SQL("{} != {}").format(column_ref, sql.Literal(str(value)))
            elif operator == FilterOperator.GT:
                return sql.SQL("({}) > {}").format(column_ref, sql.Literal(value))
            elif operator == FilterOperator.LT:
                return sql.SQL("({}) < {}").format(column_ref, sql.Literal(value))
            elif operator == FilterOperator.GTE:
                return sql.SQL("({}) >= {}").format(column_ref, sql.Literal(value))
            elif operator == FilterOperator.LTE:
                return sql.SQL("({}) <= {}").format(column_ref, sql.Literal(value))
            elif operator == FilterOperator.IN:
                if isinstance(value, list):
                    values = sql.SQL(", ").join([sql.Literal(str(v)) for v in value])
                    return sql.SQL("{} IN ({})").format(column_ref, values)
                else:
                    return sql.SQL("{} = {}").format(
                        column_ref, sql.Literal(str(value))
                    )
            elif operator == FilterOperator.NIN:
                if isinstance(value, list):
                    values = sql.SQL(", ").join([sql.Literal(str(v)) for v in value])
                    return sql.SQL("{} NOT IN ({})").format(column_ref, values)
                else:
                    return sql.SQL("{} != {}").format(
                        column_ref, sql.Literal(str(value))
                    )
            elif operator == FilterOperator.CONTAINS:
                # For JSONB array contains
                return sql.SQL("metadata -> {} ? {}").format(
                    sql.Literal(key), sql.Literal(str(value))
                )
            elif operator == FilterOperator.TEXT_MATCH:
                return sql.SQL("{} LIKE {}").format(
                    column_ref, sql.Literal(f"%{value}%")
                )
            elif operator == FilterOperator.TEXT_MATCH_INSENSITIVE:
                return sql.SQL("{} ILIKE {}").format(
                    column_ref, sql.Literal(f"%{value}%")
                )
            elif operator == FilterOperator.IS_EMPTY:
                return sql.SQL("({} IS NULL OR {} = '')").format(column_ref, column_ref)
            else:
                # Default to equality for unsupported operators
                return sql.SQL("{} = {}").format(column_ref, sql.Literal(str(value)))

        return sql.SQL("TRUE")

    filter_sqls = [_filter_to_sql(f) for f in filters.filters]

    if filters.condition.lower() == "and":
        return sql.SQL(" AND ").join(filter_sqls)
    elif filters.condition.lower() == "or":
        return sql.SQL(" OR ").join(filter_sqls)
    else:  # NOT
        if len(filter_sqls) == 1:
            return sql.SQL("NOT ({})").format(filter_sqls[0])
        else:
            return sql.SQL("NOT ({})").format(sql.SQL(" AND ").join(filter_sqls))


class AzurePGVectorStore(BasePydanticVectorStore, BaseAzurePGVectorStore):
    """Azure PostgreSQL vector store for LlamaIndex."""

    stores_text: bool = True
    metadata_columns: str | None = "metadata"

    @classmethod
    def class_name(cls) -> str:
        """Return the class name for this vector store."""
        return "AzurePGVectorStore"

    @property
    def client(self) -> None:
        """Return the client property (not used for AzurePGVectorStore)."""
        return

    @override
    @classmethod
    def from_params(
        cls,
        connection_pool: AzurePGConnectionPool,
        schema_name: str = "public",
        table_name: str = "llamaindex_vectors",
        embed_dim: int = 1536,
        embedding_index: Algorithm | None = None,
    ) -> "AzurePGVectorStore":
        """Create an AzurePGVectorStore from connection and configuration parameters."""
        return cls(
            connection_pool=connection_pool,
            schema_name=schema_name,
            table_name=table_name,
            embed_dim=embed_dim,
            embedding_index=embedding_index,
        )

    def _table_row_to_node(self, row: dict[str, Any]) -> BaseNode:
        """Convert a table row dictionary to a BaseNode object."""
        metadata = row.get("metadata")
        if metadata is None:
            raise ValueError("Metadata not found in row data.")

        node = metadata_dict_to_node(metadata, text=row.get("content"))
        # Convert UUID to string if needed
        node_id = row.get("id")
        if node_id is not None:
            node.node_id = str(node_id)
        embedding = row.get("embedding")

        if isinstance(embedding, str):
            embedding = row.get("embedding").strip("[]").split(",")
            node.embedding = list(map(float, embedding))
        elif embedding is not None:
            node.embedding = embedding
        else:
            raise ValueError("Missing embedding value")

        return node

    def _get_insert_sql_dict(
        self, node: BaseNode, on_conflict_update: bool
    ) -> tuple[sql.SQL, dict[str, Any]]:
        """Get the SQL command and dictionary for inserting a node."""
        if on_conflict_update:
            update = sql.SQL(
                """
                UPDATE SET
                    {content_col} = EXCLUDED.{content_col},
                    {embedding_col} = EXCLUDED.{embedding_col},
                    {metadata_col} = EXCLUDED.{metadata_col}
                """
            ).format(
                id_col=sql.Identifier(self.id_column),
                content_col=sql.Identifier(self.content_column),
                embedding_col=sql.Identifier(self.embedding_column),
                metadata_col=sql.Identifier(self.metadata_columns),
            )
        else:
            update = sql.SQL("nothing")
        insert_sql = sql.SQL(
            """
            INSERT INTO {schema}.{table} ({id_col}, {content_col}, {embedding_col}, {metadata_col})
            VALUES (%(id)s, %(content)s, %(embedding)s, %(metadata)s)
            ON CONFLICT ({id_col}) DO {update}
        """
        ).format(
            schema=sql.Identifier(self.schema_name),
            table=sql.Identifier(self.table_name),
            id_col=sql.Identifier(self.id_column),
            content_col=sql.Identifier(self.content_column),
            embedding_col=sql.Identifier(self.embedding_column),
            metadata_col=sql.Identifier(self.metadata_columns),
            update=update,
        )

        return (
            insert_sql,
            {
                "id": node.node_id,
                "content": node.get_content(metadata_mode=MetadataMode.NONE),
                "embedding": np.array(node.get_embedding(), dtype=np.float32),
                "metadata": Jsonb(node_to_metadata_dict(node)),
            },
        )

    @override
    def add(self, nodes: list[BaseNode], **add_kwargs: Any) -> list[str]:
        """Add a list of BaseNode objects to the vector store.

        Args:
            nodes: List of BaseNode objects to add.
            **add_kwargs: Additional keyword arguments.

        Returns:
            List of node IDs added.
        """
        ids = []
        on_conflict_update = bool(add_kwargs.pop("on_conflict_update", None))
        with self.connection_pool.connection() as conn:
            register_vector(conn)
            with conn.cursor(row_factory=dict_row) as cursor:
                for node in nodes:
                    ids.append(node.node_id)
                    insert_sql, insert_dict = self._get_insert_sql_dict(
                        node, on_conflict_update
                    )
                    cursor.execute(insert_sql, insert_dict)
        return ids

    @override
    def query(self, query: VectorStoreQuery, **kwargs: Any) -> VectorStoreQueryResult:
        """Perform a similarity search using the provided query.

        Args:
            query: VectorStoreQuery object containing the query embedding and parameters.
            **kwargs: Additional keyword arguments.

        Returns:
            VectorStoreQueryResult containing the search results.
        """
        results = self._similarity_search_by_vector_with_distance(
            embedding=query.query_embedding,
            k=query.similarity_top_k,
            filter_expression=metadata_filters_to_sql(query.filters),
            **kwargs,
        )
        nodes = []
        similarities = []
        ids = []
        for row in results:
            node = metadata_dict_to_node(row[0]["metadata"], text=row[0]["content"])
            nodes.append(node)
            similarities.append(row[1])
            ids.append(row[0]["id"])

        return VectorStoreQueryResult(
            nodes=nodes,
            similarities=similarities,
            ids=ids,
        )

    @override
    def delete(self, ref_doc_id: str, **delete_kwargs: Any) -> None:
        """Delete a node from the vector store by reference document ID.

        Args:
            ref_doc_id: The reference document ID to delete.
            **delete_kwargs: Additional keyword arguments.
        """
        with self.connection_pool.connection() as conn:
            register_vector(conn)
            with conn.cursor(row_factory=dict_row) as cursor:
                delete_sql = sql.SQL(
                    "DELETE FROM {table} WHERE metadata ->> 'doc_id' = %s"
                ).format(table=sql.Identifier(self.schema_name, self.table_name))
                cursor.execute(delete_sql, (ref_doc_id,))

    @override
    def delete_nodes(
        self,
        node_ids: Optional[list[str]] = None,
        filters: Optional[MetadataFilters] = None,
        **delete_kwargs: Any,
    ) -> None:
        """Delete nodes from the vector store by node IDs or filters.

        Args:
            node_ids: Optional list of node IDs to delete.
            filters: Optional MetadataFilters to filter nodes for deletion.
            **delete_kwargs: Additional keyword arguments.
        """
        if not node_ids:
            return

        self._delete_rows_from_table(
            ids=node_ids, filters=metadata_filters_to_sql(filters), **delete_kwargs
        )

    @override
    def clear(self) -> None:
        """Clear all data from the vector store table."""
        with self.connection_pool.connection() as conn:
            register_vector(conn)
            with conn.cursor(row_factory=dict_row) as cursor:
                stmt = sql.SQL("TRUNCATE TABLE {schema}.{table}").format(
                    schema=sql.Identifier(self.schema_name),
                    table=sql.Identifier(self.table_name),
                )
                cursor.execute(stmt)
                conn.commit()

    @override
    def get_nodes(
        self,
        node_ids: Optional[list[str]] = None,
        filters: Optional[MetadataFilters] = None,
        **kwargs: Any,
    ) -> list[BaseNode]:
        """Retrieve nodes by IDs or filters.

        Args:
            node_ids: Optional list of node IDs to retrieve.
            filters: Optional MetadataFilters to filter nodes.
            **kwargs: Additional keyword arguments.

        Returns:
            List of BaseNode objects matching the criteria.
        """
        # TODO: Implement filter handling
        documents = self._get_by_ids(node_ids)
        nodes = []
        for doc in documents:
            node = self._table_row_to_node(doc)
            nodes.append(node)

        return nodes



"""Common utilities and models for asynchronous Azure Database for PostgreSQL operations."""

from ._connection import (
    AsyncAzurePGConnectionPool,
    AsyncConnectionInfo,
    async_check_connection,
    async_create_extensions,
)

__all__ = [
    "AsyncAzurePGConnectionPool",
    "AsyncConnectionInfo",
    "async_check_connection",
    "async_create_extensions",
]



"""Synchronous connection handling for Azure Database for PostgreSQL."""

import logging
import time
from collections.abc import Callable

from azure.core.credentials import TokenCredential
from azure.identity import DefaultAzureCredential
from psycopg import Connection, sql
from psycopg.rows import dict_row
from psycopg_pool import ConnectionPool
from pydantic import ConfigDict

from ._shared import (
    TOKEN_CREDENTIAL_SCOPE,
    BaseConnectionInfo,
    BasicAuth,
    Extension,
    get_username_password,
)

_logger = logging.getLogger(__name__)


class ConnectionInfo(BaseConnectionInfo):
    """Base connection information for Azure Database for PostgreSQL connections.

    :param host: Hostname of the Azure Database for PostgreSQL server.
    :type host: str | None
    :param dbname: Name of the database to connect to.
    :type dbname: str
    :param port: Port number for the connection.
    :type port: int
    :param sslmode: SSL mode for the connection.
    :type sslmode: SSLMode
    :param credentials: Credentials for the connection.
    :type credentials: BasicAuth | TokenCredential
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True  # True to allow TokenCredential
    )

    credentials: BasicAuth | TokenCredential = DefaultAzureCredential()


def check_connection(conn: Connection, /, required_extensions: list[Extension] = []):
    """Check if the connection to Azure Database for PostgreSQL is valid and required extensions are installed.

    :param conn: Connection to the Azure Database for PostgreSQL.
    :type conn: Connection
    :param required_extensions: List of required extensions to check if they are installed.
    :type required_extensions: list[Extension]
    :raises RuntimeError: If the connection check fails or required extensions are not installed.
    """
    with conn.cursor(row_factory=dict_row) as cursor:
        _logger.debug("checking connection")
        t_start = time.perf_counter()
        cursor.execute("select 1")
        result = cursor.fetchone()
        t_elapsed = time.perf_counter() - t_start
        assert result is not None, "Connection check failed: no result returned."
        _logger.debug(
            "connection check successful. elapsed time: %.3f ms", t_elapsed * 1000
        )

        for ext in required_extensions:
            ext_name = ext.ext_name
            ext_version = ext.ext_version
            schema_name = ext.schema_name
            cursor.execute(
                sql.SQL(
                    """
                    select  extname as ext_name, extversion as ext_version,
                            n.nspname as schema_name
                      from  pg_extension e
                            left join pg_namespace n on e.extnamespace = n.oid
                     where  extname = %(ext_name)s
                    """
                ),
                {"ext_name": ext_name},
            )
            resultset = cursor.fetchone()
            if resultset is None:
                raise RuntimeError(f"Required extension '{ext_name}' is not installed.")
            if ext_version is not None and resultset["ext_version"] != ext_version:
                raise RuntimeError(
                    f"Required extension '{ext_name}' version mismatch: "
                    f"expected {ext_version}, got {resultset['ext_version']}."
                )
            if schema_name is not None and resultset["schema_name"] != schema_name:
                raise RuntimeError(
                    f"Required extension '{ext_name}' is not installed in the expected schema: "
                    f"expected {schema_name}, got {resultset['schema_name']}."
                )
            _logger.debug(
                "required extension '%s' is installed (version: %s, schema: %s)",
                resultset["ext_name"],
                resultset["ext_version"],
                resultset["schema_name"],
            )


def create_extensions(conn: Connection, /, required_extensions: list[Extension] = []):
    """Create required extensions in the Azure Database for PostgreSQL connection.

    :param conn: Connection to the Azure Database for PostgreSQL.
    :type conn: Connection
    :param required_extensions: List of required extensions to create.
    :type required_extensions: list[Extension]
    :raises Exception: If the connection is not valid or if an error occurs during extension creation.
    """
    with conn.cursor() as cursor:
        for ext in required_extensions:
            ext_name = ext.ext_name
            ext_version = ext.ext_version
            schema_name = ext.schema_name
            cascade = ext.cascade
            _logger.debug(
                "creating extension (if not exists): %s (version: %s, schema: %s, cascade: %s)",
                ext_name,
                ext_version,
                schema_name,
                cascade,
            )
            cursor.execute(
                sql.SQL(
                    """
                    create extension  if not exists {ext_name}
                                with  {schema_expr}
                                      {version_expr}
                                      {cascade_expr}
                    """
                ).format(
                    ext_name=sql.Identifier(ext_name),
                    schema_expr=sql.SQL("schema {schema_name}").format(
                        schema_name=sql.Identifier(schema_name)
                    )
                    if schema_name is not None
                    else sql.SQL(""),
                    version_expr=sql.SQL("version {version}").format(
                        version=sql.Literal(ext_version)
                    )
                    if ext_version is not None
                    else sql.SQL(""),
                    cascade_expr=sql.SQL("cascade") if cascade else sql.SQL(""),
                )
            )


class AzurePGConnectionPool(ConnectionPool):
    """Connection pool for Azure Database for PostgreSQL connections."""

    def __init__(
        self,
        conninfo: str = "",
        *,
        azure_conn_info: ConnectionInfo = ConnectionInfo(),
        **kwargs,
    ):
        if isinstance(azure_conn_info.credentials, TokenCredential):
            _logger.debug(
                "getting token from TokenCredential for the scope: %s",
                TOKEN_CREDENTIAL_SCOPE,
            )
            credential_provider = azure_conn_info.credentials
            token = credential_provider.get_token(TOKEN_CREDENTIAL_SCOPE)

            _logger.info("getting username and password from token")
            username, password = get_username_password(token)

            _logger.debug("wrapping reconnect_failed function")
            reconnect_failed: Callable[[ConnectionPool], None] | None = kwargs.get(
                "reconnect_failed"
            )

            def reconnect_failed_wrapper(pool: ConnectionPool) -> None:
                if reconnect_failed:
                    reconnect_failed(pool)

                _logger.debug(
                    "getting token from TokenCredential for the scope: %s",
                    TOKEN_CREDENTIAL_SCOPE,
                )
                token = credential_provider.get_token(TOKEN_CREDENTIAL_SCOPE)

                _logger.info("getting username and password from token")
                username, password = get_username_password(token)

                pool.kwargs.update(
                    user=username,
                    password=password,
                )

            kwargs["reconnect_failed"] = reconnect_failed_wrapper
        else:
            username, password = get_username_password(azure_conn_info.credentials)

        azure_conn_info_kwargs = azure_conn_info.model_dump(
            mode="json", exclude_none=True, exclude=set(["credentials"])
        )
        _logger.debug(
            "updating ConnectionPool kwargs with those from: %s",
            azure_conn_info_kwargs,
        )
        kwargs_ = kwargs.get("kwargs", {})
        kwargs_.update(user=username, password=password, **azure_conn_info_kwargs)
        kwargs["kwargs"] = kwargs_

        super().__init__(conninfo, **kwargs)



"""Common utilities and models for Azure Database for PostgreSQL operations."""

from ._base import (
    BaseAzurePGVectorStore,
)
from ._connection import (
    AzurePGConnectionPool,
    ConnectionInfo,
    check_connection,
    create_extensions,
)
from ._shared import (
    HNSW,
    Algorithm,
    BasicAuth,
    DiskANN,
    DiskANNIterativeScanMode,
    DiskANNSearchParams,
    Extension,
    HNSWIterativeScanMode,
    HNSWSearchParams,
    IVFFlat,
    IVFFlatIterativeScanMode,
    IVFFlatSearchParams,
    SSLMode,
    VectorOpClass,
    VectorType,
)
from .aio import (
    AsyncAzurePGConnectionPool,
    AsyncConnectionInfo,
    async_check_connection,
    async_create_extensions,
)

__all__ = [
    # Shared constructs
    "HNSW",
    "Algorithm",
    "BasicAuth",
    "DiskANN",
    "DiskANNIterativeScanMode",
    "DiskANNSearchParams",
    "Extension",
    "HNSWIterativeScanMode",
    "HNSWSearchParams",
    "IVFFlat",
    "IVFFlatIterativeScanMode",
    "IVFFlatSearchParams",
    "SSLMode",
    "VectorOpClass",
    "VectorType",
    # Base classes
    "BaseAzurePGVectorStore",
    # Synchronous connection constructs
    "AzurePGConnectionPool",
    "ConnectionInfo",
    "check_connection",
    "create_extensions",
    # Asynchronous connection constructs
    "AsyncAzurePGConnectionPool",
    "AsyncConnectionInfo",
    "async_check_connection",
    "async_create_extensions",
]
